#include "str_fields_check.h"
#include <jsonincpp/utf8.h>

bool isALPHA(char ch) {
    return ('a' <= ch && ch <= 'z') || ('A' <= ch && ch <= 'Z');
}

bool isNUM(char ch) {
    return '0' <= ch && ch <= '9';
}

bool isUNCHAR(char ch) {
    return isALPHA(ch) || isNUM(ch) || ch == '-' || ch == '_';
}

bool isSPACE(char ch) {
    return ch == ' ' || ch == '\r' || ch == '\t' || ch == '\n';
}


bool check_password(const std::string &pwd) {
    return isUtf8String(pwd) && pwd.size() >= 8;
}

bool check_name(const std::string &name) {
    return isUtf8String(name);
}

bool check_nickname(const std::string &nickname) {
    if (nickname.empty())
        return false;
    for (char ch: nickname) {
        if (!isUNCHAR(ch))
            return false;
    }
    return true;
}
